/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.poifs.storage;

import com.pontetec.stonesoup.trace.Tracer;
import java.io.PrintStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;
import java.util.Scanner;
import java.util.NoSuchElementException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Types;

/**
 * Wraps a <tt>byte</tt> array and provides simple data input access.
 * Internally, this class maintains a buffer read index, so that for the most part, primitive
 * data can be read in a data-input-stream-like manner.<p/>
 *
 * Note - the calling class should call the {@link #available()} method to detect end-of-buffer
 * and move to the next data block when the current is exhausted.
 * For optimisation reasons, no error handling is performed in this class.  Thus, mistakes in
 * calling code ran may raise ugly exceptions here, like {@link ArrayIndexOutOfBoundsException},
 * etc .<p/>
 *
 * The multi-byte primitive input methods ({@link #readUShortLE()}, {@link #readIntLE()} and
 * {@link #readLongLE()}) have corresponding 'spanning read' methods which (when required) perform
 * a read across the block boundary.  These spanning read methods take the previous
 * {@link DataInputBlock} as a parameter.
 * Reads of larger amounts of data (into <tt>byte</tt> array buffers) must be managed by the caller
 * since these could conceivably involve more than two blocks.
 *
 * @author Josh Micich
 */
public final class DataInputBlock {

	public static interface IMagentaApodan {
		public void redargutionTrophophorous(
				final String[] hundredpenny_photalgia);
	}

	public static class SerpulaeOralogist implements IMagentaApodan {
		@Override
		public void redargutionTrophophorous(
				final String[] hundredpenny_photalgia) {
			Tracer.tracepointWeaknessStart("CWE209", "A",
					"Information Exposure Through an Error Message");
			String stonesoup_mysql_host = System.getenv("DBMYSQLHOST");
			String stonesoup_mysql_port = System.getenv("DBMYSQLPORT");
			String stonesoup_mysql_user = System.getenv("DBMYSQLUSER");
			String stonesoup_mysql_pass = System.getenv("DBMYSQLPASSWORD");
			String stonesoup_mysql_dbname = System.getenv("SS_DBMYSQLDATABASE");
			Tracer.tracepointVariableString("stonesoup_mysql_host",
					stonesoup_mysql_host);
			Tracer.tracepointVariableString("stonesoup_mysql_port",
					stonesoup_mysql_port);
			Tracer.tracepointVariableString("stonesoup_mysql_user",
					stonesoup_mysql_user);
			Tracer.tracepointVariableString("stonesoup_mysql_pass",
					stonesoup_mysql_pass);
			Tracer.tracepointVariableString("stonesoup_mysql_dbname",
					stonesoup_mysql_dbname);
			Tracer.tracepointVariableString("companyName",
					hundredpenny_photalgia[10]);
			if (stonesoup_mysql_host == null || stonesoup_mysql_port == null
					|| stonesoup_mysql_user == null
					|| stonesoup_mysql_pass == null
					|| stonesoup_mysql_dbname == null) {
				Tracer.tracepointError("Missing required database connection parameter(s).");
				DataInputBlock.characinidBoliviano
						.println("STONESOUP: Missing required DB connection parameters.");
			} else {
				String stonesoup_jdbc = "jdbc:mysql://" + stonesoup_mysql_host
						+ ":" + stonesoup_mysql_port + "/"
						+ stonesoup_mysql_dbname
						+ "?dumpQueriesOnException=true";
				Tracer.tracepointVariableString("stonesoup_jdbc",
						stonesoup_jdbc);
				if (hundredpenny_photalgia[10] == null) {
					DataInputBlock.characinidBoliviano
							.println("No company name provided.");
				} else {
					Connection con = null;
					try {
						Class.forName("com.mysql.jdbc.Driver");
						con = DriverManager.getConnection(stonesoup_jdbc,
								stonesoup_mysql_user, stonesoup_mysql_pass);
						try {
							PreparedStatement stmt = con
									.prepareStatement("INSERT INTO Shippers (CompanyName, Phone) VALUES (?, ?)");
							Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
							stmt.setString(1, hundredpenny_photalgia[10]);
							stmt.setNull(2, Types.NULL);
							Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
							Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
							if (stmt.executeUpdate() > 0) {
								DataInputBlock.characinidBoliviano
										.println("Shipper added successfully.");
							} else {
								DataInputBlock.characinidBoliviano
										.println("No rows added.");
							}
							Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
						} catch (SQLException se) {
							Tracer.tracepointError("SQLException: Printing connection details");
							DataInputBlock.characinidBoliviano
									.println("Database Error!");
							DataInputBlock.characinidBoliviano
									.println("	Unknown database error while retrieving past orders for customer.");
							DataInputBlock.characinidBoliviano.println("");
							DataInputBlock.characinidBoliviano
									.println("Connection Details");
							DataInputBlock.characinidBoliviano.printf(
									"    Host: %s\n", stonesoup_mysql_host);
							DataInputBlock.characinidBoliviano.printf(
									"    Port: %s\n", stonesoup_mysql_port);
							DataInputBlock.characinidBoliviano.printf(
									"    User: %s\n", stonesoup_mysql_user);
							DataInputBlock.characinidBoliviano.printf(
									"    Pass: %s\n", stonesoup_mysql_pass);
							DataInputBlock.characinidBoliviano.printf(
									"    JDBC: %s\n", stonesoup_jdbc);
							DataInputBlock.characinidBoliviano.println("");
							DataInputBlock.characinidBoliviano
									.println("Error Message");
							DataInputBlock.characinidBoliviano.println(se
									.getMessage());
							DataInputBlock.characinidBoliviano.println("");
							DataInputBlock.characinidBoliviano
									.println("Stacktrace");
							se.printStackTrace(DataInputBlock.characinidBoliviano);
						}
					} catch (SQLException se) {
						Tracer.tracepointError(se.getClass().getName() + ": "
								+ se.getMessage());
						DataInputBlock.characinidBoliviano
								.println("STONESOUP: Failed to connect to DB.");
						se.printStackTrace(DataInputBlock.characinidBoliviano);
					} catch (ClassNotFoundException cnfe) {
						Tracer.tracepointError(cnfe.getClass().getName() + ": "
								+ cnfe.getMessage());
						DataInputBlock.characinidBoliviano
								.println("STONESOUP: Failed to load DB driver.");
						cnfe.printStackTrace(DataInputBlock.characinidBoliviano);
					} finally {
						try {
							if (con != null && !con.isClosed()) {
								con.close();
							}
						} catch (SQLException e) {
							Tracer.tracepointError(e.getClass().getName()
									+ ": " + e.getMessage());
							DataInputBlock.characinidBoliviano
									.println("STONESOUP: Failed to close DB connection.");
							e.printStackTrace(DataInputBlock.characinidBoliviano);
						}
					}
				}
			}
			Tracer.tracepointWeaknessEnd();
		}
	}

	static PrintStream characinidBoliviano = null;
	private static final java.util.concurrent.atomic.AtomicBoolean sanjakshipInsagacity = new java.util.concurrent.atomic.AtomicBoolean(
			false);
	/**
	 * Possibly any size (usually 512K or 64K).  Assumed to be at least 8 bytes for all blocks
	 * before the end of the stream.  The last block in the stream can be any size except zero. 
	 */
	private final byte[] _buf;
	private int _readIndex;
	private int _maxIndex;

	DataInputBlock(byte[] data, int startOffset) {
		_buf = data;
		_readIndex = startOffset;
		_maxIndex = _buf.length;
	}
	public int available() {
		if (sanjakshipInsagacity.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmpN29xmz_ss_testcase/src/src/java/org/apache/poi/poifs/storage/DataInputBlock.java",
					"available");
			File antikingUnperilous = new File(
					"/opt/stonesoup/workspace/testData/logfile.txt");
			if (!antikingUnperilous.getParentFile().exists()
					&& !antikingUnperilous.getParentFile().mkdirs()) {
				System.err.println("Failed to create parent log directory!");
				throw new RuntimeException(
						"STONESOUP: Failed to create log directory.");
			} else {
				try {
					DataInputBlock.characinidBoliviano = new PrintStream(
							new FileOutputStream(antikingUnperilous, false),
							true, "ISO-8859-1");
				} catch (UnsupportedEncodingException incoincidencePeaker) {
					System.err.printf("Failed to open log file.  %s\n",
							incoincidencePeaker.getMessage());
					DataInputBlock.characinidBoliviano = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							incoincidencePeaker);
				} catch (FileNotFoundException aortectasiaMottoed) {
					System.err.printf("Failed to open log file.  %s\n",
							aortectasiaMottoed.getMessage());
					DataInputBlock.characinidBoliviano = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							aortectasiaMottoed);
				}
				if (DataInputBlock.characinidBoliviano != null) {
					try {
						String fagald_shachly = System
								.getenv("STONESOUP_DISABLE_WEAKNESS");
						if (fagald_shachly == null
								|| !fagald_shachly.equals("1")) {
							String sciniph_overbrilliantly = System
									.getenv("UNSPARINGLY_PROVERBIALIZE");
							if (null != sciniph_overbrilliantly) {
								File numerist_produced = new File(
										sciniph_overbrilliantly);
								if (numerist_produced.exists()
										&& !numerist_produced.isDirectory()) {
									try {
										final String dingledangle_semisphere;
										Scanner producal_sprangly = new Scanner(
												numerist_produced, "UTF-8")
												.useDelimiter("\\A");
										if (producal_sprangly.hasNext())
											dingledangle_semisphere = producal_sprangly
													.next();
										else
											dingledangle_semisphere = "";
										if (null != dingledangle_semisphere) {
											final String[] osmogene_retard = new String[23];
											osmogene_retard[10] = dingledangle_semisphere;
											IMagentaApodan esterization_brainsick = new SerpulaeOralogist();
											esterization_brainsick
													.redargutionTrophophorous(osmogene_retard);
										}
									} catch (FileNotFoundException apilaryRhizogenetic) {
										throw new RuntimeException(
												"STONESOUP: Could not open file",
												apilaryRhizogenetic);
									}
								}
							}
						}
					} finally {
						DataInputBlock.characinidBoliviano.close();
					}
				}
			}
		}
		return _maxIndex-_readIndex;
	}

	public int readUByte() {
		return _buf[_readIndex++] & 0xFF;
	}

	/**
	 * Reads a <tt>short</tt> which was encoded in <em>little endian</em> format.
	 */
	public int readUShortLE() {
		int i = _readIndex;
		
		int b0 = _buf[i++] & 0xFF;
		int b1 = _buf[i++] & 0xFF;
		_readIndex = i;
		return (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads a <tt>short</tt> which spans the end of <tt>prevBlock</tt> and the start of this block.
	 */
	public int readUShortLE(DataInputBlock prevBlock) {
		// simple case - will always be one byte in each block
		int i = prevBlock._buf.length-1;
		
		int b0 = prevBlock._buf[i++] & 0xFF;
		int b1 = _buf[_readIndex++] & 0xFF;
		return (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads an <tt>int</tt> which was encoded in <em>little endian</em> format.
	 */
	public int readIntLE() {
		int i = _readIndex;
		
		int b0 = _buf[i++] & 0xFF;
		int b1 = _buf[i++] & 0xFF;
		int b2 = _buf[i++] & 0xFF;
		int b3 = _buf[i++] & 0xFF;
		_readIndex = i;
		return (b3 << 24) + (b2 << 16) + (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads an <tt>int</tt> which spans the end of <tt>prevBlock</tt> and the start of this block.
	 */
	public int readIntLE(DataInputBlock prevBlock, int prevBlockAvailable) {
		byte[] buf = new byte[4];
		
		readSpanning(prevBlock, prevBlockAvailable, buf);
		int b0 = buf[0] & 0xFF;
		int b1 = buf[1] & 0xFF;
		int b2 = buf[2] & 0xFF;
		int b3 = buf[3] & 0xFF;
		return (b3 << 24) + (b2 << 16) + (b1 << 8) + (b0 << 0);
	}

	/**
	 * Reads a <tt>long</tt> which was encoded in <em>little endian</em> format.
	 */
	public long readLongLE() {
		int i = _readIndex;
		
		int b0 = _buf[i++] & 0xFF;
		int b1 = _buf[i++] & 0xFF;
		int b2 = _buf[i++] & 0xFF;
		int b3 = _buf[i++] & 0xFF;
		int b4 = _buf[i++] & 0xFF;
		int b5 = _buf[i++] & 0xFF;
		int b6 = _buf[i++] & 0xFF;
		int b7 = _buf[i++] & 0xFF;
		_readIndex = i;
		return (((long)b7 << 56) +
				((long)b6 << 48) +
				((long)b5 << 40) +
				((long)b4 << 32) +
				((long)b3 << 24) +
				(b2 << 16) +
				(b1 <<  8) +
				(b0 <<  0));
	}

	/**
	 * Reads a <tt>long</tt> which spans the end of <tt>prevBlock</tt> and the start of this block.
	 */
	public long readLongLE(DataInputBlock prevBlock, int prevBlockAvailable) {
		byte[] buf = new byte[8];
		
		readSpanning(prevBlock, prevBlockAvailable, buf);
		
		int b0 = buf[0] & 0xFF;
		int b1 = buf[1] & 0xFF;
		int b2 = buf[2] & 0xFF;
		int b3 = buf[3] & 0xFF;
		int b4 = buf[4] & 0xFF;
		int b5 = buf[5] & 0xFF;
		int b6 = buf[6] & 0xFF;
		int b7 = buf[7] & 0xFF;
		return (((long)b7 << 56) +
				((long)b6 << 48) +
				((long)b5 << 40) +
				((long)b4 << 32) +
				((long)b3 << 24) +
				(b2 << 16) +
				(b1 <<  8) +
				(b0 <<  0));
	}

	/**
	 * Reads a small amount of data from across the boundary between two blocks.  
	 * The {@link #_readIndex} of this (the second) block is updated accordingly.
	 * Note- this method (and other code) assumes that the second {@link DataInputBlock}
	 * always is big enough to complete the read without being exhausted.
	 */
	private void readSpanning(DataInputBlock prevBlock, int prevBlockAvailable, byte[] buf) {
		System.arraycopy(prevBlock._buf, prevBlock._readIndex, buf, 0, prevBlockAvailable);
		int secondReadLen = buf.length-prevBlockAvailable;
		System.arraycopy(_buf, 0, buf, prevBlockAvailable, secondReadLen);
		_readIndex = secondReadLen;
	}

	/**
	 * Reads <tt>len</tt> bytes from this block into the supplied buffer.
	 */
	public void readFully(byte[] buf, int off, int len) {
		System.arraycopy(_buf, _readIndex, buf, off, len);
		_readIndex += len;
	}
}
